"""Counts the number of lines, words, and characters"""

label = "Word count"
modes = ["ide"]
tooltip = "Counts the number of lines, words, and characters"
shortcut = "Ctrl+Alt+C"
icon = "edit-find"
priority = -2
