"""Quickly jump to functions, classes, and other symbol definitions"""

tooltip = "Quickly jump to functions, classes, and other symbol definitions"
label = "Symbol selector"
shortcut = "Ctrl+R"
icon = "edit-find"
modes = ["ide"]
settings = {
    "symbol_selector_shortcut": "Ctrl+R"
}
