"""Add images as annotations to code"""

label = "Image annotations"
modes = ["ide"]
settings = {
    "image_annotations": "{}",
    "image_annotations_enabled": True,
    "image_annotations_capture_output": False
}
