"""Find text in project files"""

label = "Find in project files"
tooltip = "Find text in project files"
shortcut = "Ctrl+Shift+F"
icon = "edit-find"
menu = {
    "index": 0,
    "submenu": "Edit"
}
modes = ["ide"]
